@@ -21,6 +21,12 @@ from dataclasses import dataclass
 from types import ModuleType
 from typing import Any, Callable, Dict, List, Optional, Type, Union
 
+# For Python 3.7
+try:
+    from typing import Literal  # type: ignore
+except ImportError:
+    from typing_extensions import Literal  # type: ignore
+
 import lark
 import rich
 from enforce_typing import enforce_types
@@ -41,19 +47,11 @@ class _Ast(ast_utils.Ast):
     pass
 
 
-class _Statement(_Ast):
-    @enforce_types
-    def assign_field(self, clazz: Type, values: List[Any]):
-        for v in values:
-            if isinstance(v, clazz):
-                return v
-
-
 @dataclass
 class Value(_Ast):
     value: str
 
-    def __str__(self):
+    def __str__(self) -> str:
         return str(self.value)
 
 
@@ -61,7 +59,7 @@ class Value(_Ast):
 class Name(_Ast, ast_utils.AsList):
     name: List[str]
 
-    def __str__(self):
+    def __str__(self) -> str:
         return ".".join([str(x) for x in self.name])
 
 
@@ -69,7 +67,7 @@ class Name(_Ast, ast_utils.AsList):
 class Expression(_Ast, ast_utils.AsList):
     statements: List[_Ast]
 
-    def __str__(self):
+    def __str__(self) -> str:
         msg = ""
         for s in self.statements:
             msg += f"{s}"
@@ -77,32 +75,33 @@ class Expression(_Ast, ast_utils.AsList):
         return msg
 
 
+@dataclass
 class _JoinType(_Ast):
-    def __str__(self):
+    def __str__(self) -> str:
         return "JOIN"
 
 
 @dataclass
 class InnerJoin(_JoinType):
-    def __str__(self):
+    def __str__(self) -> Literal["INNER JOIN"]:
         return "INNER JOIN"
 
 
 @dataclass
 class LeftJoin(_JoinType):
-    def __str__(self):
+    def __str__(self) -> Literal["LEFT JOIN"]:
         return "LEFT JOIN"
 
 
 @dataclass
 class RightJoin(_JoinType):
-    def __str__(self):
+    def __str__(self) -> Literal["RIGHT JOIN"]:
         return "RIGHT JOIN"
 
 
 @dataclass
 class OuterJoin(_JoinType):
-    def __str__(self):
+    def __str__(self) -> Literal["OUTER JOIN"]:
         return "OUTER JOIN"
 
 
@@ -110,7 +109,7 @@ class OuterJoin(_JoinType):
 class JoinType(_JoinType):
     join_type: _JoinType
 
-    def __str__(self):
+    def __str__(self) -> str:
         if self.join_type is None:
             return "JOIN"
         else:
@@ -124,34 +123,17 @@ class Join(_Ast):
     left_id: Optional[Name] = None  # Has to have a default argument now
     right_id: Optional[Name] = None
 
-    def __init__(
-        self,
-        name: Name,
-        join_type: Optional[_JoinType] = None,
-        left_id: Name = None,
-        right_id: Optional[Name] = None,
-    ):
-        self.name = name
-        self.join_type = join_type
-
+    def __post_init__(self) -> None:
         if isinstance(self.name, JoinType):
             temp = self.join_type
             self.join_type = self.name
             self.name = temp  # type: ignore[assignment]
-
         if isinstance(self.join_type, Name):
             # Now we need to shift everything , since join_type is now our left_id
-
-            temp = left_id
+            temp = self.left_id
             self.left_id = self.join_type
             self.right_id = temp
             self.join_type = None
-        else:
-            self.left_id = left_id
-            self.right_id = right_id
-
-        # self.left_id = left_id
-        # self.right_id = right_id
 
 
 @dataclass
@@ -159,7 +141,7 @@ class SelectField(_Ast):
     name: Name
     cast_type: Optional[Name] = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         if self.cast_type is not None:
             return f"CAST({self.name} as {self.cast_type})"
         return str(self.name)
@@ -169,15 +151,15 @@ class SelectField(_Ast):
 class SelectFields(_Ast, ast_utils.AsList):
     fields: List[SelectField]
 
-    def __str__(self):
-        return [str(x) for x in self.fields]
+    def __str__(self) -> str:
+        return ",".join([str(x) for x in self.fields])
 
 
 @dataclass()
 class SortField(_Ast):
     name: Name
 
-    def __str__(self):
+    def __str__(self) -> str:
         return str(self.name)
 
 
@@ -185,7 +167,7 @@ class SortField(_Ast):
 class SortFields(_Ast, ast_utils.AsList):
     fields: List[SortField]
 
-    def __str__(self):
+    def __str__(self) -> str:
         return ",".join([str(x) for x in self.fields])
 
 
@@ -193,7 +175,7 @@ class SortFields(_Ast, ast_utils.AsList):
 class PipeBody(_Ast):
     body: str
 
-    def __str__(self):
+    def __str__(self) -> str:
         return str(self.body)
 
 
@@ -201,7 +183,7 @@ class PipeBody(_Ast):
 class Select(_Ast):
     fields: SelectFields
 
-    def __str__(self):
+    def __str__(self) -> str:
         return ",".join([str(f) for f in self.fields.fields])
 
 
@@ -209,7 +191,7 @@ class Select(_Ast):
 class DeriveBody(_Ast):
     val: Union[str, Expression]
 
-    def __str__(self):
+    def __str__(self) -> str:
         return str(self.val)
 
 
@@ -217,7 +199,7 @@ class DeriveBody(_Ast):
 class Operator(_Ast):
     op: Token
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.op.value
 
 
@@ -225,7 +207,7 @@ class Operator(_Ast):
 class GroupBy(_Ast, ast_utils.AsList):
     fields: List[str]
 
-    def __str__(self):
+    def __str__(self) -> str:
         return ",".join([str(f) for f in self.fields])
 
 
@@ -233,27 +215,22 @@ class GroupBy(_Ast, ast_utils.AsList):
 class AggregateBody(_Ast, ast_utils.AsList):
     statements: List[_Ast]
 
-    def __str__(self):
+    def __str__(self) -> str:
         return f"{[str(s) for s in self.statements]}"
 
 
 @dataclass
-class Aggregate(_Statement):
-    group_by: GroupBy
+class Aggregate(_Ast):
+    group_by: Optional[GroupBy] = None
     aggregate_body: Optional[AggregateBody] = None
 
-    def __init__(self, group_by, aggregate_body=None):
-        values = [group_by, aggregate_body]
-        self.aggregate_body = self.assign_field(AggregateBody, values)
-        self.group_by = self.assign_field(GroupBy, values)
-
 
 @dataclass
 class DeriveLine(_Ast):
     name: str
     expression: Expression
 
-    def __str__(self):
+    def __str__(self) -> str:
         return str(self.expression)
 
 
@@ -270,7 +247,7 @@ class _Direction(_Ast):
 class Direction(_Ast):
     direction: Optional[_Direction] = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         return str(self.direction)
 
 
@@ -278,7 +255,7 @@ class Direction(_Ast):
 class Ascending(_Direction):
     direction = "ASC"
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.direction
 
 
@@ -286,7 +263,7 @@ class Ascending(_Direction):
 class Descending(_Direction):
     direction = "DESC"
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.direction
 
 
@@ -295,7 +272,7 @@ class Sort(_Ast):
     fields: SortFields
     direction: Optional[Direction] = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         return (
             f'{str(self.fields)} {self.direction if self.direction is not None else ""}'
         )
@@ -306,7 +283,7 @@ class Take(_Ast):
     qty: str
     offset: Optional[str] = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         sql = f"LIMIT {self.qty}"
         if (
             self.offset is not None
@@ -327,7 +304,7 @@ class Filter(_Ast, ast_utils.AsList):
 class FilterLine(_Ast):
     val: Any = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         return str(self.val)
 
 
@@ -337,26 +314,16 @@ class Pipes(_Ast, ast_utils.AsList):
 
 
 @dataclass
-class From(_Statement):
-    name: str
+class From(_Ast):
+    name: Optional[Name] = None
     pipes: Optional[Pipes] = None
 
-    def __init__(self, name, pipes=None, join=None):
-        self.name = name
-        values = [pipes, join]
-
-        self.pipes = self.assign_field(Pipes, values)
-        self.join = self.assign_field(Join, values)
-
-    def get_pipes(self):
-        return self.pipes
-
 
 @dataclass
 class FuncArgs(_Ast, ast_utils.AsList):
     fields: Optional[List] = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         return ",".join([str(x) for x in self.fields])
 
 
@@ -366,17 +333,11 @@ class FuncBody(_Ast, ast_utils.AsList):
 
 
 @dataclass
-class FuncDef(_Statement):
+class FuncDef(_Ast):
     name: Name
-    func_args: FuncArgs
+    func_args: Optional[FuncArgs] = None
     func_body: Optional[FuncBody] = None
 
-    def __init__(self, name, func_args=None, func_body=None):
-        self.name = name
-        values = [func_args, func_body]
-        self.func_args = self.assign_field(FuncArgs, values)
-        self.func_body = self.assign_field(FuncBody, values)
-
 
 @dataclass
 class FuncCall(_Ast):
@@ -415,26 +376,34 @@ class WithDef(_Ast):
 
 
 # The top level definition that holds all other definitions
-@dataclass
-class Root(_Statement):
-    with_def: Optional[WithDef] = None
-    _from: Optional[From] = None
-    value_defs: Optional[ValueDefs] = None
-    func_defs: Optional[FuncDefs] = None
-
-    def __init__(self, with_def=None, _from=None, value_def=None, func_def=None):
+class Root(_Ast):
+    def __init__(
+        self,
+        with_def: Optional[WithDef] = None,
+        _from: Optional[From] = None,
+        value_def: Optional[ValueDefs] = None,
+        func_def: Optional[FuncDefs] = None,
+    ) -> None:
         values = [with_def, _from, value_def, func_def]
         self.with_def = self.assign_field(WithDef, values)
         self._from = self.assign_field(From, values)
         self.value_defs = self.assign_field(ValueDefs, values)
         self.func_defs = self.assign_field(FuncDefs, values)
 
-    def get_from(self):
+    def get_from(self) -> Optional[From]:
         return self._from
 
-    def get_cte(self):
+    def get_cte(self) -> Optional[WithDef]:
         return self.with_def
 
+    @enforce_types
+    def assign_field(self, clazz: Type, values: List[Any]) -> Optional[Type]:
+        for v in values:
+            if isinstance(v, clazz):
+                return v
+        # Explicit statement of prior implicit behaviour to silence MyPy
+        return None
+
 
 # This is a helper class, its not actually parsed
 @dataclass
@@ -442,22 +411,22 @@ class NameValuePair(_Ast):
     name: str
     value: str
 
-    def __str__(self):
+    def __str__(self) -> str:
         return f"({self.value})"
 
 
 class ToAst(Transformer):
-    def STRING(self, s):
+    def STRING(self, s: str) -> str:
         # Remove quotation marks
         return s  # s[1:-1]
 
-    def SSTRING(self, s):
+    def SSTRING(self, s: str) -> str:
         return s[2:-1]
 
-    def ESCAPED_STRING(self, s):
+    def ESCAPED_STRING(self, s: str) -> str:
         return s  # s.replace('\\"', '"').replace("\\'", "'")
 
-    def NEWLINE(self, s):
+    def NEWLINE(self, s: str) -> str:
         return s
 
 
@@ -469,14 +438,7 @@ def read_file(filename: str, path: str = script_path) -> str:
 
 
 @enforce_types
-def get_func_str(func: Optional[str]) -> str:
-    if func == "average":
-        return "avg"
-    return func if isinstance(func, str) else str(func)
-
-
-@enforce_types
-def parse(_text: str) -> Root:
+def parse(_text: str, verbose: bool = False) -> Root:
     global GLOBAL_PARSER
     global GLOBAL_TRANSFORMER
     text = _text + "\n"
@@ -488,7 +450,8 @@ def parse(_text: str) -> Root:
             transformer=ToAst(),
         )
     tree = GLOBAL_PARSER.parse(text)
-
+    if verbose:
+        rich.print(tree)
     if GLOBAL_TRANSFORMER is None:
         GLOBAL_TRANSFORMER = ast_utils.create_transformer(this_module, ToAst())
     return GLOBAL_TRANSFORMER.transform(tree)
@@ -497,11 +460,12 @@ def parse(_text: str) -> Root:
 @enforce_types
 def to_sql(prql: str, verbose: bool = False) -> str:
     global STDLIB_AST
-    ast = parse(prql)
+    ast = parse(prql, verbose)
     if verbose:
         rich.print(ast)
     if STDLIB_AST is None:
         STDLIB_AST = parse(read_file("stdlib.prql"))
+    # @formatter:off
     return (
         ast_to_sql(ast.get_from(), [ast, STDLIB_AST], verbose=verbose)
         .replace("   ", " ")
@@ -560,14 +524,14 @@ def get_operation(
 
 
 @enforce_types
-def alias(s: str, n: int = 1):
+def alias(s: str, n: int = 1) -> str:
     return s + "_" + s[0:n]
 
 
 @enforce_types
 def replace_all_tables_standalone(
     from_long: str, from_short: str, join_long: List[str], join_short: List[str], s: str
-):
+) -> str:
     s = s.replace(f"{from_long}.", f"{from_short}.")
     if join_long and len(join_long) > 0:
         for i in range(len(join_long)):
@@ -578,8 +542,8 @@ def replace_all_tables_standalone(
 @enforce_types
 def wrap_replace_all_tables(
     from_long: str, from_short: str, join_long: List[str], join_short: List[str]
-) -> Callable:
-    def inner(x):
+) -> Callable[[str], str]:
+    def inner(x: str) -> str:
         return replace_all_tables_standalone(
             from_long, from_short, join_long, join_short, x
         )
@@ -600,8 +564,8 @@ def replace_tables_standalone(
 @enforce_types
 def wrap_replace_tables(
     from_long: str, from_short: str, join_long: str, join_short: str
-) -> Callable:
-    def inner(x):
+) -> Callable[[str], str]:
+    def inner(x: str) -> str:
         return replace_tables_standalone(
             from_long, from_short, join_long, join_short, x
         )
@@ -618,9 +582,9 @@ def build_symbol_table(roots: List[Root]) -> Dict[str, List[_Ast]]:
             table[str(n.name)].append(n)
         for n in root.func_defs.fields:
             table[str(n.name)].append(n)
-        if root.get_from().get_pipes():
+        if root.get_from().pipes:
             derives = get_operation(
-                root.get_from().get_pipes().operations, Derive, return_all=True
+                root.get_from().pipes.operations, Derive, return_all=True
             )
             for d in derives:
                 for line in d.fields:
@@ -693,6 +657,14 @@ def execute_function(
             for line in func_def.func_body.fields:
                 if isinstance(line, PipeBody):
                     line = line.body
+                if isinstance(line, Expression):
+                    exp: Expression = line
+                    if len(exp.statements) != 1:
+                        raise PRQLException(
+                            "Currently only single statement functions that are text based are supported"
+                        )
+                    else:
+                        line = str(exp.statements[0])
                 if type(line) == str:
                     args = {}
 
@@ -758,13 +730,20 @@ def safe_to_sql(
     roots: Union[Root, List],
     symbol_table: Optional[Dict] = None,  # Dict[str, List[_Ast]]
     verbose: bool = False,
-):
+) -> Optional[str]:
+    # Return must be Optional[str] as no else clause is provided.
+    # In python, this makes the check non-exhaustive, so the implicit None
+    # return is a possibility.
     if isinstance(rule, str):
         return rule
     elif isinstance(rule, Token) or isinstance(rule, Name):
         return str(rule)
     elif isinstance(rule, _Ast):
         return ast_to_sql(rule, roots, symbol_table, verbose)
+    # I've added the else clause explicitly to quiet mypy.
+    # This is just an explicit statement of previous behaviour.
+    else:
+        return None
 
 
 @enforce_types
@@ -773,7 +752,7 @@ def ast_to_sql(
     roots: Union[Root, List],
     symbol_table: Optional[Dict] = None,  # Optional[Dict[str, List[_Ast]]]
     verbose: bool = True,
-):
+) -> str:
     if isinstance(roots, Root):
         root = roots
     else:
@@ -804,7 +783,7 @@ def ast_to_sql(
         ###
 
         _from = rule
-        ops = _from.get_pipes()
+        ops = _from.pipes
         joins: List[Join] = get_operation(ops.operations, Join, return_all=True)
         agg: Aggregate = get_operation(ops.operations, Aggregate)
         take: Take = get_operation(ops.operations, Take, last_match=True)
